import sasktranif.sasktranif as skif
import numpy as np
import sasktran as sk
import logging
from sasktran.exceptions import wrap_skif_functionfail
from collections import namedtuple
from sasktran.util import DictWithCallback, LowercaseKeyDictWithCallback, to_iter


class Engine(object):
    """
    Base class for objects which can perform radative transfer calculations.
    """

    @wrap_skif_functionfail
    def __init__(self, name: str, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        """ Base class constructor. The base class cannot be instantiated.
        """
        if name.lower() == 'hr' and not isinstance(self, EngineHR):
            raise ValueError("Using sasktran.Engine('hr') is no longer supported, use sasktran.EngineHR() instead ")

        self._name = name
        self._iskengine = skif.ISKEngine(self._name)
        self._wavelengths = wavelengths
        if geometry is not None:
            self._geometry = geometry
        else:
            self._geometry = sk.Geometry()

        if atmosphere is not None:
            self._atmosphere = atmosphere
        else:
            self._atmosphere = sk.Atmosphere()
        self._needs_reconfigure = True
        self._options = LowercaseKeyDictWithCallback(options) if options is not None else LowercaseKeyDictWithCallback()
        self._options.add_modified_callback(self._set_needs_reconfigure)

    def __getstate__(self):
        state = {k: v for k, v in self.__dict__.items() if k != '_iskengine'}
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._iskengine = skif.ISKEngine(self._name)
        self._needs_reconfigure = True

    def skif_object(self, **kwargs):
        """ Get the underlying ISK object.

        Returns
        -------
        skif.ISKEngine
            Underlying ISKEngine object
        """
        return self._iskengine

    @property
    def options(self):
        """
        dict : Dictionary map of additional engine options.

        Examples
        --------
        >>> import sasktran as sk
        >>> engine = sk.EngineHR()
        >>> engine.options['numordersofscatter'] = 1
        """
        return self._options

    @property
    def atmosphere(self):
        """
        sasktran.Atmosphere : The engines atmosphere definition object.
        """
        return self._atmosphere

    @atmosphere.setter
    def atmosphere(self, value):
        self._atmosphere = value
        self._needs_reconfigure = True

    @property
    def geometry(self):
        """
        sasktran.Geometry : The engines geometry definition object.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        self._geometry = value
        self._needs_reconfigure = True

    @property
    def wavelengths(self):
        """
        np.ndarray or float : Wavelengths in nm to perform the radiative transfer calculation at.

        Examples
        --------
        >>> import sasktran as sk
        >>> engine = sk.EngineHR()
        >>> engine.wavelengths = 600 # Single wavelength
        >>> engine.wavelengths = [600, 340] # Multiple wavelengths
        """
        return self._wavelengths

    @wavelengths.setter
    def wavelengths(self, value):
        self._wavelengths = value

    @wrap_skif_functionfail
    def calculate_radiance(self, output_format='numpy'):
        """ Calculate radiances.

        Parameters
        ----------
        output_format : str, optional
            Determines the output format, one of ['numpy' or 'xarray'].  If output_format='numpy', the radiance
            and optionally the weighting functions are returned as numpy arrays.  If output_format='xarray' the
            radiance and weighting functions are returned inside of an xarray Dataset object.

        Returns
        -------
        np.ndarray **or** np.ndarray, np.ndarray **or** xarray.Dataset
            If output_format='numpy':
            The first np.ndarray is always the computed radiances. The shape of the computed
            radiances is (W, L) where W is the number of wavelengths, and L is the number of
            lines of sight.
            If ``atmosphere.wf_species != None`` then a second np.ndarray will be returned
            containing the computed weighting functions. The shape of this np.ndarray is
            engine specific but typically (W, L, A) where A is the number of altitudes at which
            at which the weighting functions were calculated.

            If output_format='xarray':
            A dataset containing the radiance and weighting functions with appropriate coordinates.
        """
        if self._needs_reconfigure or self._atmosphere.has_changed or self._geometry.has_changed:
            self._iskengine = skif.ISKEngine(self._name)
            self._add_atmosphere_wf()
            self._add_lines_of_sight()
            self._add_options()
            self._initialize_model()
            self._add_atmosphere()
            self._needs_reconfigure = False
            self._geometry._added_to_engine()
            self._atmosphere._added_to_engine()

        if self._wavelengths is not None:
            self._iskengine.SetWavelengths(np.atleast_1d(self._wavelengths))
        else:
            raise ValueError("Property wavelengths needs to be set before calling Engine.calculate_radiance()")

        data = self._iskengine.CalculateRadiance()[1]

        if output_format.lower() == 'numpy':
            return data
        elif output_format.lower() == 'xarray':
            try:
                import xarray as xr
            except ModuleNotFoundError:
                raise ValueError('Output Format "xarray" requires the package xarray to be installed')

            los_vectors = np.vstack([l.look_vector for l in self.geometry.lines_of_sight])
            obs_positions = np.vstack([l.observer for l in self.geometry.lines_of_sight])
            mjds = np.vstack([l.mjd for l in self.geometry.lines_of_sight]).flatten()

            data = xr.Dataset({'radiance': (['wavelength', 'los'], data),
                               'mjd': (['los'], mjds),
                               'los_vector': (['los', 'xyz'], los_vectors),
                               'observer_position': (['los', 'xyz'], obs_positions)},
                              coords={'wavelength': self.wavelengths,
                                      'xyz': ['x', 'y', 'z']})

            return data
        else:
            raise ValueError('Supported values for output_format are "numpy" and "xarray"')

    @wrap_skif_functionfail
    def _initialize_model(self):
        self._iskengine.InitializeModel()

    def _set_needs_reconfigure(self):
        self._needs_reconfigure = True

    @wrap_skif_functionfail
    def _add_lines_of_sight(self):
        for los in to_iter(self._geometry.lines_of_sight):
            self._iskengine.AddLineOfSight(los.mjd, los.observer, los.look_vector)

        if self._geometry.sun is not None:
            self.options['setsun'] = self._geometry.sun

    @wrap_skif_functionfail
    def _add_atmosphere(self):
        for _, species in self._atmosphere.species.items():
            self._iskengine.AddSpecies(species.skif_species, species.climatology.skif_object(engine=self, opt_prop=species.optical_property),
                                       species.optical_property.skif_object(engine=self))
        if self._atmosphere.atmospheric_state is not None:
            self._iskengine.SetAtmosphericState(self._atmosphere.atmospheric_state.skif_object())
        if self._atmosphere.brdf is not None:
            self._iskengine.SetBRDF(self._atmosphere.brdf.skif_object(engine=self))

    @wrap_skif_functionfail
    def _add_atmosphere_wf(self):
        if self._atmosphere.wf_species is not None:
            if self._atmosphere.wf_species in self._atmosphere.species:
                op = self._atmosphere.species[self._atmosphere.wf_species].optical_property.skif_object()
                self._iskengine.SetProperty('wfspecies', op)
                if 'calcwf' not in [key.lower() for key in self._options]:
                    self._iskengine.SetProperty('calcwf', 2)
            else:
                raise ValueError("WFspecies '%s' not found in species list" % self._atmosphere.wf_species)

    @wrap_skif_functionfail
    def _add_options(self):
        for key, val in self._options.items():
            self._iskengine.SetProperty(key, val)

    @wrap_skif_functionfail
    def _reset_engine(self):
        self._iskengine = skif.ISKEngine(self._name)

        self._needs_reconfigure = True


class EngineHR(Engine):
    """
    Examples
    --------
    >>> import sasktran as sk
    >>> # configure your geometry
    >>> geometry = sk.VerticalImage()
    >>> geometry.from_sza_saa(60, 60, 0, 0, [10, 20, 30, 40], 54372, 0)
    >>> # configure your atmosphere
    >>> atmosphere = sk.Atmosphere()
    >>> atmosphere['rayleigh'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
    >>> atmosphere['o3'] = sk.Species(sk.O3DBM(), sk.Labow())
    >>> atmosphere['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.Pratmo())
    >>> atmosphere.brdf = sk.Lambertian(0.3)
    >>> # make your engine
    >>> engine = sk.EngineHR(geometry=geometry, atmosphere=atmosphere, wavelengths=[350, 400, 450])
    >>> rad = engine.calculate_radiance()
    >>> print(rad) #doctest: +ELLIPSIS
    [[ 0.11674...  0.10872...  0.04904...  0.01456...]
     [ 0.10837...  0.08537...  0.02897...  0.00804...]
     [ 0.10318...  0.06396...  0.01829...  0.00483...]]
    """
    def __init__(self, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        super().__init__('hr', geometry, atmosphere, wavelengths, options)

        self._model_parameters = dict()

    def _initialize_model(self):
        super()._initialize_model()

        self._model_parameters['referencepoint'] = self._iskengine.GetProperty('ReferencePoint')[1]
        self._model_parameters['sun'] = self._iskengine.GetProperty('sun')[1]

    def _add_atmosphere(self):
        super()._add_atmosphere()

        # Add in any atmospheric emissions
        for species, emission in self._atmosphere.emissions.items():
            if emission.skif_species_override():
                guid = emission.skif_species_override()
            else:
                guid = species
                skif.AddGeneratedGlobalClimatologyHandleIfNotExists(guid)
            self._iskengine.AddEmission(guid, emission.skif_object(atmosphere=self._atmosphere, engine=self))

    def _add_lines_of_sight(self):
        super()._add_lines_of_sight()

        if self._geometry.reference_point is not None:
            self._iskengine.SetProperty('setreferencepoint', self._geometry.reference_point)

    @property
    def model_parameters(self):
        return self._model_parameters

    @wrap_skif_functionfail
    def calculate_radiance(self, output_format='numpy'):
        rad = super().calculate_radiance(output_format)
        if self._atmosphere.wf_species is not None:
            wf = self._iskengine.GetWeightingFunctions()[1]
            if output_format.lower() == 'numpy':
                output = namedtuple('EngineOutput', ['radiance', 'weighting_function'])
                return output(radiance=rad, weighting_function=wf)
            elif output_format.lower() == 'xarray':
                if self.options.get('calcwf', 2) == 2:
                    rad['wf_' + self.atmosphere.wf_species] = (['wavelength', 'los', 'altitude'], wf)
                    rad.coords['altitude'] = ('altitude', self.options.get('wfheights', np.linspace(500, 99500, 100)))
                    rad = rad.stack(perturbation=('altitude',))
                else:
                    rad['wf_' + self.atmosphere.wf_species] = (['wavelength', 'los', 'perturbation'], wf)

                return rad
        else:
            return rad

    @property
    def num_orders_of_scatter(self):
        """
        Controls the number of orders of scatter used inside the model. The number of orders of scatter controls the
        accuracy of the model.  When set to 1, only light directly scattered from the sun is accounted for, with no
        multiple scattering.

        The model runs substantially faster when set
        to 1 scatter order, therefore a common use case is to do development/testing with this property set to 1.  There
        is very little speed difference between 2 orders of scatter and 50 orders of scatter, so usually it is
        recommended to set this to either 1 or 50.

        **Default:** 50
        """
        return self.options.get('numordersofscatter', 50)

    @num_orders_of_scatter.setter
    def num_orders_of_scatter(self, num_orders_of_scatter: int):
        if num_orders_of_scatter < 1 or not isinstance(num_orders_of_scatter, int):
            raise ValueError('num_orders_of_scatter must be a positive integer')
        self.options['numordersofscatter'] = num_orders_of_scatter

    @property
    def surface_elevation(self):
        """
        Sets the elevation of the surface in [m] for the model.  This setting can be used to account for changes in the Earth's
        topography.

        **Default:** 0 m
        """
        return self.options.get('surfaceheight', 0)

    @surface_elevation.setter
    def surface_elevation(self, surface_height: float):
        self.options['surfaceheight'] = surface_height

    @property
    def top_of_atmosphere_altitude(self):
        """
        Sets the altitude of the top of the atmosphere.  Above this altitude it is assumed there is no atmosphere.
        The default value of 100000 m is typically suitable for stratospheric applications but for the mesosphere it
        is sometimes necessary to increase this value.

        **Default:** 100000 m
        """
        return self.options.get('toaheight', 100000)

    @top_of_atmosphere_altitude.setter
    def top_of_atmosphere_altitude(self, toa_altitude: float):
        self.options['toaheight'] = toa_altitude

    @property
    def num_threads(self):
        """
        Controls the number of threads to use when multithreading the radiative transfer calculation.  The default value
        of 0 indicates that the number of threads used will be equal to the number of available logical cores on
        the target machine.

        Setting this value to a lower number than the number of available cores can be useful when running a radiative
        transfer calculation in the background and the computer is too slow to multitask.

        **Default:** 0
        """
        return self.options.get('numthreads', 0)

    @num_threads.setter
    def num_threads(self, num_threads: int):
        if num_threads < 0 or not isinstance(num_threads, int):
            raise ValueError('num_threads must be an integer >= 0')
        self.options['numthreads'] = num_threads

    @property
    def grid_spacing(self):
        """
        Sets many options dealing with general grids and discretizations in the model.  In general, if you are interested
        in things on the scale of 1000 m, then set this to 1000 m.  For most applications setting this property is
        sufficient, however for advanced users it may be necessary to set some grids individually.


        **Default:** 1000 m
        """
        return self.options.get('raytracingshells', 1000)

    @grid_spacing.setter
    def grid_spacing(self, grid: float):
        if grid < 0:
            raise ValueError('Grid Spacing should be positive')

        self.options['raytracingshells'] = grid
        self.options['solarraytracingshells'] = grid

        optical_heights = np.arange(self.surface_elevation, self.top_of_atmosphere_altitude, grid / 2)
        optical_heights = np.concatenate((optical_heights, [self.top_of_atmosphere_altitude]))

        self.options['manualopticalheights'] = optical_heights

        diffuse_heights = np.arange(self.surface_elevation + 1e-6, self.top_of_atmosphere_altitude, grid * 5)

        self.options['manualdiffuseheights'] = diffuse_heights

    @property
    def include_emissions(self):
        """
        Enables support for atmospheric emissions inside the model.  If set to True, any :py:class:`Emission <sasktran.Emission>`
        objects that are included in the :py:class:`Atmosphere <sasktran.Atmosphere>` will be used in the radiative
        transfer calculation.  Note that the default value is False, ignoring emissions.

        **Default:** False
        """
        return bool(self.options.get('useemissions', False))

    @include_emissions.setter
    def include_emissions(self, use_emissions: bool):
        self.options['useemissions'] = int(use_emissions)

    @property
    def atmosphere_dimensions(self):
        """
        Sets the number of dimensions for the atmosphere in the radiative transfer calculation.


        ========== ======================================
        Input      Setting
        ========== ======================================
        1          One-dimensional in altitude
        2          Two-dimensional in altitude and angle
                   along the line of sight
        3          Fully three-dimensional
        ========== ======================================

        **Default:** 1
        """
        opt_type = self.options.get('OpticalTableType', 0)

        if opt_type == 0:
            return 1
        if opt_type == 1:
            return 3
        if opt_type == 2:
            return 2

    @atmosphere_dimensions.setter
    def atmosphere_dimensions(self, atmosphere_dim: int):
        if atmosphere_dim < 0 or atmosphere_dim > 3:
            raise ValueError('atmosphere_dim must be either 1, 2, or 3')
        if atmosphere_dim == 1:
            self.options['opticaltabletype'] = 0
        if atmosphere_dim == 2:
            self.options['opticaltabletype'] = 2
        if atmosphere_dim == 3:
            self.options['opticaltabletype'] = 1

    @property
    def num_diffuse_profiles(self):
        """
        Advanced option that controls the number of diffuse profiles in the model.  If worried about the accuracy
        of the multiple scatter signal, this should be the first option to try changing.  A diffuse profile
        represents a (latitude, longitude) where the multiple scatter signal is calculated.  The default value of 1
        indicates that the multiple scatter signal is only calculated at the reference point, or the average tangent
        point.  If you increase the value to 5, then there will be 5 locations along the line of sight where the full
        multiple scatter calculation is done.  Outside of these areas the multiple scatter signal is interpolated. Odd
        numbers of diffuse profiles are preferred (but not necessary) since this facilitates one profile being placed
        at the tangent point.

        To determine if this setting has an effect on your calculation it is recommended to first compare the results
        between 1, 5, and 11 diffuse profiles.

        The amount of diffuse profiles to obtain an accurate depends heavily on a variety of things, but most heavily
        the solar zenith and solar scattering angles.  When the sun is high in the sky (low solar zenith angle) 1--3
        diffuse profiles is usually sufficient.  However, when looking across the solar terminator (sza ~90, ssa < 90)
        it might be necessary to use upwards of 50 diffuse profiles to obtain a signal accurate to better than 0.5%.
        More information on the number of required diffuse profiles can be found in
        Zawada, D. J., Dueck, S. R., Rieger, L. A., Bourassa, A. E., Lloyd, N. D., and Degenstein, D. A.: High-resolution and Monte Carlo additions to the SASKTRAN radiative transfer model, Atmos. Meas. Tech., 8, 2609-2623, https://doi.org/10.5194/amt-8-2609-2015, 2015.
        """
        return self.options.get('numdiffuseprofilesinplane', 1)

    @num_diffuse_profiles.setter
    def num_diffuse_profiles(self, num_diffuse: int):
        if num_diffuse < 1 or not isinstance(num_diffuse, int):
            raise ValueError('num_diffuse must be an integer greater than 0')

        self.options['numdiffuseprofilesinplane'] = num_diffuse


class EngineDO(Engine):
    """
    Examples
    --------
    >>> import sasktran as sk
    >>> # Configure the geometry
    >>> geometry = sk.NadirGeometry()
    >>> tempo = sk.Geodetic()
    >>> tempo.from_lat_lon_alt(0, -100, 35786000)
    >>> lats = np.linspace(20, 50, 4)
    >>> lons = np.linspace(-110, -90, 3)
    >>> lats, lons = np.meshgrid(lats, lons, indexing='ij')
    >>> geometry.from_lat_lon(mjd=57906.63472, observer=tempo, lats=lats, lons=lons)
    >>> # Configure the atmosphere
    >>> atmosphere = sk.Atmosphere()
    >>> atmosphere['rayleigh'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
    >>> atmosphere['o3'] = sk.Species(sk.O3DBM(), sk.Labow())
    >>> atmosphere['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.Pratmo())
    >>> atmosphere.brdf = sk.Kokhanovsky()
    >>> # Calcualte radiance
    >>> engine = sk.EngineDO(geometry=geometry, atmosphere=atmosphere, wavelengths=[350, 450])
    >>> rad = engine.calculate_radiance()
    >>>
    >>> # print radiances
    >>> print(rad[0].reshape((4, 3))) # radiance at 350nm
    [[ 0.15726175  0.19468353  0.22767423]
     [ 0.1666762   0.20027906  0.22962238]
     [ 0.17116222  0.1995542   0.22411801]
     [ 0.17099479  0.1932349   0.21235274]]
    >>> print(rad[1].reshape((4, 3))) # radiance at 450nm
    [[ 0.15884236  0.19815234  0.23287579]
     [ 0.16822132  0.20341264  0.23415734]
     [ 0.17226753  0.20184478  0.22736965]
     [ 0.1716201   0.19441591  0.21402527]]
    """
    def __init__(self, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        super().__init__('do', geometry, atmosphere, wavelengths, options)
        self._model_parameters = dict()
        self._wf_shape = None
        self._options = LowercaseKeyDictWithCallback(options) if options is not None else LowercaseKeyDictWithCallback()
        self._options.add_modified_callback(self._set_needs_reconfigure)

    def _initialize_model(self):
        super()._initialize_model()
        self._model_parameters['sun'] = self._iskengine.GetProperty('sun')[1]
        self.model_parameters['numstreams'] = int(self._iskengine.GetProperty('numstreams')[1])
        self.model_parameters['numlayers'] = int(self._iskengine.GetProperty('numlayers')[1])
        self.model_parameters['numphasefexpansionterms'] = \
            int(self._iskengine.GetProperty('numphasefexpansionterms')[1])
        self.model_parameters['numbrdfexpansionterms'] = int(self._iskengine.GetProperty('numbrdfexpansionterms')[1])

    @wrap_skif_functionfail
    def _add_atmosphere_wf(self):
        if self.atmosphere.wf_species is not None:
            wf_species_handles = []
            for a_species in np.atleast_1d(self.atmosphere.wf_species):
                if not (a_species in self.atmosphere.species):
                    err_msg = "Weighting function species '{}' was not in the species list".format(a_species)
                    raise ValueError(err_msg)
                wf_species_handles.append(self.atmosphere.species[a_species].species)
            self.options['wfclimatologies'] = " ".join(wf_species_handles)
            if not ('wfaltitudes' in self.options):
                self.options['wfaltitudes'] = np.linspace(500, 99500, 100)
            if not ('wfwidths' in self.options):
                self.options['wfwidths'] = np.array([1000] * 100)
            if hasattr(self.atmosphere.wf_species, '__iter__') and not isinstance(self.atmosphere.wf_species, str):
                self._wf_shape = (
                    len(self.wavelengths),
                    len(self.geometry.lines_of_sight),
                    len(self.atmosphere.wf_species),
                    len(self.options['wfaltitudes'])
                )
            else:
                self._wf_shape = (
                    len(self.wavelengths),
                    len(self.geometry.lines_of_sight),
                    len(self.options['wfaltitudes'])
                )

    @property
    def model_parameters(self):
        return self._model_parameters

    @wrap_skif_functionfail
    def _add_lines_of_sight(self):
        for los in to_iter(self._geometry.lines_of_sight):
            self._iskengine.AddLineOfSight(los.mjd, los.observer, los.look_vector)

        if self._geometry.sun is not None:
            self.options['sun'] = self._geometry.sun
        if self._geometry.reference_point is not None:
            logging.warning('the reference point set in the geometry object is ignored by EngineDO')

    @wrap_skif_functionfail
    def calculate_radiance(self, output_format='numpy'):
        rad = super().calculate_radiance(output_format)
        if self._atmosphere.wf_species is not None:
            wf = self._iskengine.GetWeightingFunctions()[1].reshape(self._wf_shape)
            if output_format.lower() == 'numpy':
                output = namedtuple('EngineOutput', ['radiance', 'weighting_function'])
                return output(radiance=rad, weighting_function=wf)
            elif output_format.lower() == 'xarray':
                if hasattr(self.atmosphere.wf_species, '__iter__') and not isinstance(self.atmosphere.wf_species, str):
                    for idx, species in enumerate(self.atmosphere.wf_species):
                        rad['wf_' + species] = (['wavelength', 'los', 'perturbation'], wf[:, :, idx, :])
                else:
                    rad['wf_' + self.atmosphere.wf_species] = (['wavelength', 'los', 'perturbation'], wf)
                rad.coords['altitude'] = ('altitude', self.options.get('wfaltitudes', np.linspace(500, 99500, 100)))
                rad = rad.stack(perturbation=('altitude',))
                return rad
        else:
            return rad

    @wrap_skif_functionfail
    def get_reference_points(self):
        reference_points = np.ndarray((len(self._geometry.lines_of_sight), 4))
        for i, los in enumerate(self._geometry.lines_of_sight):
            ok, ref_pt = self._iskengine.GetProperty('referencepoint[{}]'.format(i))
            reference_points[i, :] = ref_pt
        return reference_points

    @property
    def num_streams(self):
        return self.model_parameters.get('numstreams')

    @num_streams.setter
    def num_streams(self, n: int):
        self.options['numstreams'] = n

    @property
    def num_layers(self):
        return self.model_parameters.get('numlayers')

    @num_layers.setter
    def num_layers(self, n: int):
        self.options['numlayers'] = n

    @property
    def num_phasef_quadrature_terms(self):
        return self.model_parameters.get('numphasefexpansionterms')

    @num_phasef_quadrature_terms.setter
    def num_phasef_quadrature_terms(self, n: int):
        self.options['numphasefexpansionterms'] = n

    @property
    def num_brdf_quadrature_terms(self):
        return self.model_parameters.get('numbrdfexpansionterms')

    @num_brdf_quadrature_terms.setter
    def num_brdf_quadrature_terms(self, n: int):
        self.options['numbrdfexpansionterms'] = n